"""
Script to rewrite DB DNA files in a more legible format

Example usage:
>>> dna = Genome('mybot.txt')
>>> print dna
"""

import re

rStop = re.compile(r'stop.*?(cond|start|else|$)')
rEmpty = re.compile(r'(else|start) (cond|else|start)')
rCond = re.compile(r'cond (.*?) (start|stop|else)')
rStart = re.compile(r'start (.*?) (cond|stop|else)')
rElse = re.compile(r'else (.*?) (cond|stop|start)')

logic  = ['not', 'and', 'or', 'xor']
comp   =  ['=', '!=', '<', '>', '>=', '<=', '%=', '!%=', '~=', '!~=']
stores = ['inc', 'dec', 'store']
def stackInc(num):
    return num + 1
def stackDec(num):
    return num - 1
def negate(num):
    return -num
def complement(num):
    return ~num
def sgn(num):
    if num>0:
        return 1
    elif num < 0:
        return -1
    else:
        return 0
unaryOps = {'++': stackInc, '--': stackDec, '-': negate, '~': complement,
            'abs': abs, 'sgn': sgn, 'sqr': lambda x:x*x}
unhandledUnary = ['rnd']
def modulo(x, y):
    pass
div = lambda x,y: int(round(float(x)/y)) if y else 0
binaryOps = {'add': lambda x,y: x + y,  'sub':lambda x,y: x - y,
             'mult':lambda x,y: x * y,  'div': div,
             '&':   lambda x,y: x & y,  '^':  lambda x,y: x ^ y,
             '|':   lambda x,y: x | y,
             '>>':  lambda x,y: x >>y,  '<<': lambda x,y: x << y,
             'ceil': min,               'floor': max}
unhandledBinary = ['dist', 'angle', 'mod', 'pyth']
associative = ['add', 'mult', '&', '^', '|', 'ceil', 'floor']
unaryInverse = {'++': '--', '--': '++', '-': '-', '~': '~'}

def approxEqu(x, y):
    return abs(x-y) <= div(x, 10)

binComp = {'=': lambda x,y: x == y, '!=': lambda x,y: x != y,
           '<': lambda x,y: x < y,  '>': lambda x,y: x > y,
           '>=':lambda x,y: x >= y, '<=': lambda x,y: x <= y,
           '%=': approxEqu, '!%=': lambda x,y: not approxEqu(x,y)}
binLogic = {'and': lambda x, y: x and y, 'or': lambda x,y: x or y,
            'xor': lambda x,y: x == y}

arity = {}
for op in unaryOps.keys():
    arity[op] = 1
for op in unhandledUnary:
    arity[op] = 1
for op in binaryOps.keys():
    arity[op] = 2
for op in unhandledBinary:
    arity[op] = 2
arity['inc'] = arity['dec'] = 1
arity['store'] = 2
for op in binComp.keys():
    arity[op] = 2
for op in binLogic.keys():
    arity[op] = 2
arity['not'] = 1

def isNumber(block):
    try:
        return isinstance(eval(block), int)
    except:
        return False
def isBool(block):
    try:
        return isinstance(eval(block), bool)
    except:
        return False

def reduceOps(commands):
    """
    Simplifies a series of DB DNA commands.
    """
    i = 0
    while i < len(commands) - 1:
        bp0 = commands[i]
        bp1 = commands[i+1]
        bp2 = commands[i+2] if i+2 < len(commands) else None
        bpp = commands[i-1] if i>0 else None                 

        if bp0 == '-' and bp1 == '~': #special case : ~-x == x-1
            commands[i:i+2] = ['--']
            i = i - 1 if i > 0 else 0
        elif bp0 == '~' and bp1 == '-': #special case : -~x == x+1
            commands[i:i+2] = ['++']
            i = i - 1 if i > 0 else 0
        elif bp0 in unaryInverse.keys() and bp1 == unaryInverse[bp0]:
            commands[i:i+2] = []
            i = i - 2 if i > 1 else 0
        elif bp1 in logic and bp0 not in (logic + comp):
            commands[i:i+2] = [bp1, bp0]
            i = i - 1 if i > 0 else 0
        elif (isNumber(bp0) or bp0.startswith(('.','*')) ) and \
             bp1 in ['inc', 'dec'] and bpp  \
             and not bpp.startswith('*') \
             and not bpp in stores:
            commands[i-1:i+2] = [bp0, bp1, bpp]
            i -= 1
        elif isNumber(bp0):
            if bp1 in unaryOps.keys():
                commands[i:i+2] = [str(unaryOps[bp1](eval(bp0)))]
                i = i - 1 if i > 0 else 0
            elif bp0 == '0' and bp1 in ['inc', 'dec']:
                commands[i:i+2] = []
                i = i - 2 if i > 1 else 0
            elif bp2 and isNumber(bp1) and bp2 in binaryOps.keys():
                commands[i:i+3] = [str(binaryOps[bp2](eval(bp0), eval(bp1)))]
                i = i - 1 if i > 0 else 0
            else:
                i+=1
        else:
            i+=1

class Expression(object):
    def __init__(self, head, args = None):
        if isinstance(head, Expression):
            self.head = head.head
            self.args = head.args[:]
        elif not args is None:
            self.head = head
            self.args = [Expression(arg) for arg in args]           
        else:
            if isNumber(head) or isBool(head):
                self.head = eval(head)
                self.args = []
            elif head.startswith(('*','.')):
                self.head = head
                self.args = []
            elif head == 'dup':
                self.head = head
                self.args = []
            else:
                self.head = head
                self.args = [None for i in range(arity[head])]

    def __repr__(self):
        return "Expression(" + `self.head` + "," + `self.args` + ")"

    def needArgs(self):
        return (None in self.args)

    def getArg(self, exprList, default = '0'):
        if None not in self.args:
            raise ValueError
        for i in reversed(range(len(self.args))):
            if self.args[i] is None:
                if exprList:
                    self.args[i] = exprList.pop()
                else:
                    self.args[i] = Expression(default)
                if self.args[i].head in stores:
                    self.args.insert(i, None)
    def simplify(self):
        for arg in self.args:
            if not (isinstance(arg.head,int) or isinstance(arg.head,bool)):
                break
        else:
            self.compute()

    def compute(self):
        if self.head in associative:
            self.head = reduce(binaryOps[self.head], [arg.head for arg in self.args])
            self.args = []
        elif self.head in binaryOps.keys():
            self.head = binaryOps[self.head](self.args[0].head, self.args[1].head)
            self.args = []
        elif self.head in unaryOps.keys():
            self.head = unaryOps[self.head](self.args[0].head)
            self.args = []
        elif self.head == 'not':
            self.head = not self.args[0].head
            self.args = []
        elif self.head in binLogic.keys():
            self.head = binLogic[self.head](self.args[0].head, self.args[1].head)
            self.args = []
        elif self.head in binComp.keys():
            self.head = binComp[self.head](self.args[0].head, self.args[1].head)
            self.args = []
    def __str__(self):
        return ' '.join([str(arg) for arg in (self.args + [self.head])])
            
class Parser(object):
    def __init__(self):
        self.cond = []
        self.main = []

    def parse(self, token):
        expr = Expression(token)
        while expr.needArgs():
            if token in (logic):
                expr.getArg(self.cond, 'True')
            else:
                expr.getArg(self.main)
        expr.simplify()
        if token in (logic + comp):
            self.cond.append(expr)
        else:
            self.main.append(expr)
        print token, "parsed"

    def __str__(self):
        string = '\n'.join([str(expr) for expr in self.cond])
        string = string + '\n' + '\n'.join([str(expr) for expr in self.main])
        return string
        

class Block(object):
    def __init__(self, kind, content):
        self.kind = kind
        if isinstance(content, str):
            self.content = content.split()
        else:
            self.content = content[:]
        self.clean()

    def __repr__(self):
        return "Block( " + repr(self.kind) + ", " + repr(self.content) + ")"

    def __eq__(self, other):
        if isinstance(other, Block):
            return self.kind == other.kind and self.content == other.content
        else:
            return false
        
    def __add__(self, other):
        return Block.join(self, other)

    def asList(self):
        return [self.kind] + self.content[:]

    def clean(self):
        if self.kind == 'cond':
            for token in stores:
                while token in self.content:
                    self.content.remove(token)
        elif self.kind in ['start','else']:
            for token in (logic + comp):
                while token in self.content:
                    self.content.remove(token)
        elif self.kind == 'stop':
            self.content = []
        else:
            raise ValueError, "Invalid block encountered"
       
    def join(self, other):
        if self.kind != other.kind:
            raise ValueError, "Only blocks of the same kind can be joined"
        return Block(self.kind, self.content + other.content)

class Gene(object):
    def __init__(self, blocks):
        self.cond = []
        self.body = []
        self.elsebody = []
        kindlist = [b.kind for b in blocks]
        if 'cond' in kindlist and 'cond' in kindlist[1:]:
            raise ValueError, "Invalid gene"
        if blocks[0].kind == 'cond':
            self.cond = blocks[0].content[:]
        startBlocks = filter(lambda x: x.kind == 'start', blocks)
        if startBlocks:
            self.body = reduce(Block.join, startBlocks).content[:]
        elseBlocks = filter(lambda x: x.kind == 'else', blocks)
        if self.cond and elseBlocks:
            self.elsebody = reduce(Block.join, elseBlocks).content[:]
        self.simplify()

    def tokens(self):
        tok = ['cond'] + self.cond[:]
        if self.body:
            tok += ['start'] + self.body
        if self.elsebody:
            tok += ['else'] + self.elsebody
        if not (self.body or self.elsebody): #need a delimiter for end of cond block
            tok += ['stop']
        return tok

    def simplify(self):
        for w in (logic + comp):
            if w in self.cond:
                break
        else:   #cond stack is never touched, gene is always true
            self.body = self.cond + self.body
            self.cond = []
            self.elsebody=[]
        reduceOps(self.cond)
        reduceOps(self.body)
        reduceOps(self.elsebody)
    def append(self, other):
        if other.cond or other.elsebody:
            raise ValueError
        self.body = self.body + other.body
        self.simplify()

    def parseTree(self):
        self.condparser = Parser()
        for tok in self.cond:
            self.condparser.parse(tok)
        self.bodyparser = Parser()
        for tok in self.body:
            self.bodyparser.parse(tok)
        self.elseparser = Parser()
        for tok in self.elsebody:
            self.elseparser.parse(tok)
        
        

def format(dna):
        work = dna
        work = re.sub(r'(?<=\s)(inc|dec|store|not|and|or|xor|=|!=|<|>|>=|<=|%=|!%=|~=|!~=)\s+', r'\1\n ', work)
        work = re.sub(r'\b(cond|start|else|stop)\s+', r'\1\n ', work)
        work = re.sub(r'\s+(not|and|or|xor)\b', r'\n \1', work)
        work = re.sub(r'\s+(start|else|stop)\b', r'\n\1', work)
        work = re.sub(r'\s+(cond)\b', r'\n\n\1', work)
        return work

class Genome(object):
    def __init__(self,filename):
        self.filename = filename
        self.raw = file(filename).read()
        self.tokens, self.genes = self.parse()

    def __str__(self):
        return format(' '.join(self.tokens))

    def parse(self):
        work = re.sub(r"'.*",'',self.raw)
        work = ' '.join(work.split())
        work = 'stop '+ work +' stop'
        rBlocks = re.compile(r'\b(cond|start|else|stop)\b(.*?)\b(?=cond|start|else|stop)\b')
        blocks = []
        for m in rBlocks.finditer(work):
            blocks.append(Block(m.group(1),m.group(2)))

        # Merge consecutive cond blocks (and the others as well)
        i = 0
        while i < len(blocks) - 1:
            if blocks[i].kind == blocks[i+1].kind:
                blocks[i:i+2] = [blocks[i] + blocks[i+1]]
            else:
                i += 1
        blocks = filter(lambda x: x.kind != 'stop', blocks) #rm stop blocks
        geneBreaks = []
        for i, block in enumerate(blocks):
            if block.kind == 'cond' and i != 0:
                geneBreaks.append(i)
        genes = [blocks[:]]
        for i in reversed(geneBreaks):
            genes[0:1] = [genes[0][:i], genes[0][i:]]
        for i, gene in enumerate(genes):
            genes[i] = Gene(gene)
        while len(genes) > 1 and not genes[0].cond and not genes[1].cond:
            genes[0].append(genes[1])
            genes[0:2] = [genes[0]]
        tokens = reduce(lambda x,y: x+y, [g.tokens() for g in genes])
        tokens.reverse()
        cut = len(tokens)
        for w in stores:
            if w in tokens:
                cut = min(cut, tokens.index(w))
        tokens = tokens[cut:]
        tokens.reverse()
        return tokens, genes

    def clean(self):
        work = re.sub(r"'.*",'',self.raw)
        work = ' '.join(work.split())
        work = 'stop '+ work +' stop'
        work = re.sub(rStop, r"stop \1", work)
        m = rCond.search(work)
        while m:
            block = m.group(0)
            for w in [' cond ',' inc ', ' dec ', ' store ' ]:
                block = block.replace(w,' ')
            block = ' '.join(block.split())
            pos = m.start() + len(block)
            work = work[:m.start()]+ block + work[m.end():]
            m = rCond.search(work, pos)
        m = rStart.search(work)
        while m:
            block = m.group(0)
            for w in [' start ', ' not ', ' and ', ' or ', ' xor ', ' = ', ' != ', ' < ', ' > ',\
                      ' >= ', ' <= ', ' %= ', ' !%= ', ' ~= ', ' !~= ']: 
                block = block.replace(w,' ')
            pos = m.start() + len(block)
            work = work[:m.start()] + block + work[m.end():]
            m = rStart.search(work, pos)
        m = rElse.search(work)
        while m:
            block = m.group(0)
            for w in [' else ', ' not ', ' and ', ' or ', ' xor ', ' = ', ' != ', ' < ', ' > ',\
                      ' >= ', ' <= ', ' %= ', ' !%= ', ' ~= ', ' !~= ']: 
                block = block.replace(w,' ')
            pos = m.start() + len(block)
            work = work[:m.start()] + block + work[m.end():]
            m = rStart.search(work, pos)
        #work = ' '.join(work.split())
        work = work[5:]     #strip initial 'stop '
        work = re.sub(r'(?<=\s)(cond|start|else|stop|inc|dec|store|not|and|or|xor|=|!=|<|>|>=|<=|%=|!%=|~=|!~=)\s+', r'\1\n ', work)
        work = re.sub(r'\s+(not|and|or|xor)\b', r'\n \1', work)
        work = re.sub(r'\s+(start|else|stop)\b', r'\n\1', work)
        work = re.sub(r'\s+(cond)\b', r'\n\n\1', work)
        return work

